#ifndef _COMMON_SEQ_H
#define _COMMON_SEQ_H

#include <vector>
#include <iostream>
#include <cmath>

#include "util.h"

// Procedures to find common subsequences are implemented below from
// Lodhi et al. (2002) and Rousu & Shawe Taylor (2005)

extern double LAMBDA; 

using namespace std;

template <typename Term>
void rowncs(int start, const Term &x, const Term &y, int N, vector <double> &r) { 
   // finds common subsequences between x[start:end] and y and stores it in r[end]

   //vector <double> r(x.size(), 0);
   r.resize(x.size(), 0);

   double lambda=LAMBDA; // lambda==1 will give the number of common subsequences
   double lambdaSq = lambda*lambda;

   if (x.size()-start < N) N = x.size()-start;
   if (y.size() < N) N = y.size();
   
   vector <vector <double> > k(x.size()-start, vector <double> (y.size(), 0));
   vector <vector <double> > cs(x.size()-start, vector <double> (N+1, 0));
   
   // length one
   for (int i=start; i<x.size(); ++i) {
      for (int j=0; j<y.size(); ++j) {
	 if (x[i]==y[j]) {
	    k[i-start][j] = lambdaSq;
	    for (int g=i-start; g<cs.size(); ++g) cs[g][1] += k[i-start][j];
	 }
      }
   }
   for (int g=0; g<x.size()-start; ++g) cs[g][1] /= lambdaSq;

   // > 1 lengths
   for (int l=2; l<=N; ++l) {
      vector <vector <double> > S(x.size()+1-start, vector <double> (y.size()+1, 0));

      for (int i=start; i<x.size(); ++i) {
         for (int j=0; j<y.size(); ++j) {
	    S[i+1-start][j+1] = k[i-start][j] + lambda*S[i-start][j+1] + lambda*S[i+1-start][j] - lambdaSq*S[i-start][j];
	    if (x[i]==y[j]) {
	       k[i-start][j] = lambdaSq*S[i-start][j];
	       for (int g=i-start; g<cs.size(); ++g) cs[g][l] += k[i-start][j];
	    }
	 }
      }
      double pl = power(lambda, 2*l);
      for (int g=0; g<x.size()-start; ++g) cs[g][l] /= pl;
   }
      
   for (int g=0; g<x.size()-start; ++g) {
      for (int l=1; l<=N; ++l) {
         r[g] += cs[g][l];
      }
   }
}


template <typename Term>
void selfrowncs(int start, const Term &x, int N, vector <double> &r) { 
    // finds common subsequences between x[start:end] and x and stores it in r[end]

   //vector <double> r(x.size(), 0);
   r.resize(x.size(), 0);

   double lambda=LAMBDA; // lambda==1 will give the number of common subsequences
   double lambdaSq = lambda*lambda;

   if (x.size()-start < N) N = x.size()-start;
   
   vector <vector <double> > k(x.size()-start, vector <double> (x.size()-start, 0));
   vector <vector <double> > cs(x.size()-start, vector <double> (N+1, 0));
   
   // length one
   for (int i=start; i<x.size(); ++i) {
      for (int j=start; j<x.size(); ++j) {
	 if (x[i]==x[j]) {
	    k[i-start][j-start] = lambdaSq;
	       
	    for (int g=i-start; g<cs.size(); ++g) 
	       if (g>=j-start) cs[g][1] += k[i-start][j-start];
	 }
      }
   }
   for (int g=0; g<x.size()-start; ++g) cs[g][1] /= lambdaSq;

   // > 1 lengths
   for (int l=2; l<=N; ++l) {
      vector <vector <double> > S(x.size()+1-start, vector <double> (x.size()+1-start, 0));

      for (int i=start; i<x.size(); ++i) {
         for (int j=start; j<x.size(); ++j) {
	    S[i+1-start][j+1-start] = k[i-start][j-start] + lambda*S[i-start][j+1-start] + lambda*S[i+1-start][j-start] - lambdaSq*S[i-start][j-start];
	    if (x[i]==x[j]) {
	       k[i-start][j-start] = lambdaSq*S[i-start][j-start];
	       for (int g=i-start; g<cs.size(); ++g) 
		   if (g>=j-start) cs[g][l] += k[i-start][j-start];
	    }
	 }
      }
      double pl = power(lambda, 2*l);
      for (int g=0; g<x.size()-start; ++g) cs[g][l] /= pl;
   }
      
   for (int g=0; g<x.size()-start; ++g) {
      for (int l=1; l<=N; ++l) {
         r[g] += cs[g][l];
      }
   }
}

template <typename Term>
void allncs_helper(const Term &x, const Term &y, int N, vector <vector <double> > &r) { 
   r.resize(x.size());
   for (int i=0; i<x.size(); ++i) {
      rowncs(i,x,y,N,r[i]);
   }
}

template <typename Term>
void selfncs_helper(const Term &x, int N, vector <vector <double> > &r) { 
   r.resize(x.size());
   for (int i=0; i<x.size(); ++i) {
      selfrowncs(i,x,N,r[i]);
   }
}

template <typename Term>
void allncs(const Term &x, const Term &y, int N, vector <vector <double> > &r) { 
   r.resize(x.size());
   allncs_helper(x,y,N,r);
   vector <vector <double> > rx(x.size());
   selfncs_helper(x,N,rx);
   double yn = ncs_helper(y,y,N);
   for (int i=0; i<x.size(); ++i) {
      for (int j=0; j<x.size()-i; ++j) {
	 r[i][j] /= sqrt(rx[i][j]*yn);
      }
   }
}


template <typename Term>
void allncsUn(const Term &x, const Term &y, int N, vector <vector <double> > &r) { 
   r.resize(x.size());
   allncs_helper(x,y,N,r);
   vector <vector <double> > rx(x.size());
   selfncs_helper(x,N,rx);
   double yn = ncs_helper(y,y,N);
   /*
   for (int i=0; i<x.size(); ++i) {
      for (int j=0; j<x.size()-i; ++j) {
	 r[i][j] /= sqrt(rx[i][j]*yn);
      }
   }*/
}



template <typename Term>
double ncs_helper(const Term &x, const Term &y, int N) { 

   double r=0;

   double lambda=LAMBDA; // lambda==1 will give the number of common subsequences
   double lambdaSq = lambda*lambda;

   if (x.size() < N) N = x.size();
   if (y.size() < N) N = y.size();
   
   vector <vector <double> > k(x.size(), vector <double> (y.size(), 0));
   vector <double> cs(N+1, 0);

   // length one
   cs[1] = 0;
   for (int i=0; i<x.size(); ++i) {
      for (int j=0; j<y.size(); ++j) {
	 if (x[i]==y[j]) {
	    k[i][j] = lambdaSq;
	    cs[1] += k[i][j];
	 }
      }
   }
   cs[1] /= lambdaSq;

   // > 1 lengths
   for (int l=2; l<=N; ++l) {
      vector <vector <double> > S(x.size()+1, vector <double> (y.size()+1, 0));
      cs[l] = 0;

      for (int i=0; i<x.size(); ++i) {
         for (int j=0; j<y.size(); ++j) {
	    S[i+1][j+1] = k[i][j] + lambda*S[i][j+1] + lambda*S[i+1][j] - lambdaSq*S[i][j];
	    if (x[i]==y[j]) {
	       k[i][j] = lambdaSq*S[i][j];
	       cs[l] += k[i][j];
	    }
	 }
      }
      cs[l] /= power(lambda, 2*l);
   }
   for (int l=1; l<=N; ++l) {
      r += cs[l];
   }
   return r;
}

template <typename Term>
double ncs_helper(const Term &x, int xs, int xe, const Term &y, int ys, int ye, int N) { 
   double r=0;

   double lambda=LAMBDA; // lambda==1 will give the number of common subsequences
   double lambdaSq = lambda*lambda;

   int xsize = xe-xs+1;
   int ysize = ye-ys+1;

   if (xsize < N) N = xsize;
   if (ysize < N) N = ysize;
   
   vector <vector <double> > k(xsize, vector <double> (ysize, 0));
   vector <double> cs(N+1, 0);

   // length one
   cs[1] = 0;
   for (int i=xs; i<=xe; ++i) {
      int ix = i-xs;
      for (int j=ys; j<=ye; ++j) {
	 int jy = j-ys;
	 if (x[i]==y[j]) {
	    k[ix][jy] = lambdaSq;
	    cs[1] += k[ix][jy];
	 }
      }
   }
   cs[1] /= lambdaSq;

   // > 1 lengths
   for (int l=2; l<=N; ++l) {
      vector <vector <double> > S(xsize+1, vector <double> (ysize+1, 0));
      cs[l] = 0;

      for (int i=xs; i<=xe; ++i) {
	 int ix = i-xs;
         for (int j=ys; j<=ye; ++j) {
	    int jy = j-ys;
	    S[ix+1][jy+1] = k[ix][jy] + lambda*S[ix][jy+1] + lambda*S[ix+1][jy] - lambdaSq*S[ix][jy];
	    if (x[i]==y[j]) {
	       k[ix][jy] = lambdaSq*S[ix][jy];
	       cs[l] += k[ix][jy];
	    }
	 }
      }
      cs[l] /= power(lambda, 2*l);
   }
   for (int l=1; l<=N; ++l) {
      r += cs[l];
   }
   return r;
}

template <typename Term>
double ncsUn(const Term &x, const Term &y, int N) { 
   return ncs_helper(x,y,N);
}

template <typename Term>
double ncsUn(const Term &x, int xs, int xe, const Term &y, int ys, int ye, int N) { 
   return ncs_helper(x,xs,xe,y,ys,ye,N);
}

template <typename Term>
double ncs(const Term &x, const Term &y, int N) { 
   return ncs_helper(x,y,N)/sqrt(ncs_helper(x,x,N)*ncs_helper(y,y,N));
}

template <typename Term>
double ncs(const Term &x, int xs, int xe, const Term &y, int ys, int ye, int N) { 
   return ncs_helper(x,xs,xe,y,ys,ye,N)/sqrt(ncs_helper(x,xs,xe,x,xs,xe,N)*ncs_helper(y,ys,ye,y,ys,ye,N));
}

#endif
