#include <iostream>
#include <fstream>
#include <cassert>
#include <string>
#include <map>
#include <functional>
#include <algorithm>

#include "util.h"
#include "parse.h"
#include "production.h"
#include "example.h"

using namespace std;

string START;

// parameters
int KERNEL; // type of kernel: string, tree or combined
double TREE_LAMBDA=0.2; // tree kernel parameter
int STR_KERNEL_LEN=3; // string kernel parameter

int MAX_MEM=1000000; // maximum storage to avoid some recomputations (memory vs time trade-off)
int MAX_MEM2=10000; 


double LAMBDA=0.25; // string kernel parameter



vector <production> prods; // all the productions
vector <vector <string> > synProds; // syntactic productions

// maps to make string computations efficient 
map <string, int> s2n; // mapping from strings to numbers
vector <string> n2s; // reverse mapping
map <string, int> nonterm; // map from non-terminals to indices


void readGr(const string &file) {
   // read formal grammar
   ifstream in(file.c_str());

   map<string, int>::iterator itr;

   nonterm[""] = 0;

   
   string s = "";
   while (in && s != "<start>") {
      in>>s;
   }
   assert(in);
   in>>START;
   in>>s;
   assert(s == "</start>");

   
   while (in) {
      production tmp;
      if (tmp.read(in)) {

	 itr = nonterm.find(tmp.getLhs());
	 if (itr==nonterm.end()) {
	    int si = nonterm.size();
	    nonterm[tmp.getLhs()] = si;
	 }

	 if (tmp.isConst) {
	    tmp.pn = nonterm[tmp.getLhs()];
	 }
	 else {
	    tmp.pn = prods.size();
	 }
	  if (tmp.pat.size() > 0) {
            tmp.isConst = true;
            tmp.learnable = false;
         }
         else {
            tmp.isConst = false;
            tmp.learnable = true;
         }
	 prods.push_back(tmp);
	 
      }
   }
   in.close();

   // if more than two productions with the same constant pattern 
   for (int i=0; i<prods.size(); ++i) {
      prods[i].dupl = false;
      for (int j=0; j<prods.size(); ++j) {
	 if (i != j) {
	    if (prods[i].pat==prods[j].pat) { // later take reorders into account
	       prods[i].dupl = true;
	       break;
	    }
	 }
      }
   }
}

void readTrainExamples(vector <example> &trainEg, const string &filename, const mrlParser &mp) {
   ifstream in(filename.c_str());
   assert(in);

  // read nl and clang examples
   bool b = true;
   while (b) {
      int id;
      vector<std::string> nl1, nl;
      vector <clang> p1;
      double conf;
      b = readExample(in, id, nl1, p1, conf, mp);

      if (b) {
         nGramize(nl1, nl, 1);
	 trainEg.push_back(example(nl, p1, id, prods.size(), conf));
      }
   }
   in.close();
}

void test(const  std::string &cfile, const std::string &ifile, const std::string &ofile, const mrlParser &mp) {
   vector <example> ceg, ieg; // correct and input examples

   readTrainExamples(ceg, cfile,mp);
   readTrainExamples(ieg, ifile,mp);

   for (int i=0; i<ieg.size(); ++i) {
      ieg[i].eval = true;
   }
   sort(ieg.begin(), ieg.end());

   vector <bool> correct(ieg.size(), false);

   for (int i=0; i<ieg.size(); ++i) {
      for (int j=0; j<ceg.size(); ++j) {
	 if (ieg[i].id==ceg[j].id) {
	    if (ieg[i].c==ceg[j].c) {
	       correct[i] = true;
	       break;
	    }
	 }
      }
   }

   int tp=0;
   vector <double> p, r;

   int att=0, corr=0, incorr=0;

   ofstream out(ofile.c_str());
   
   for (int i=0; i<ieg.size(); ++i) {
      out<<ieg[i]<<"\n";
      if (correct[i]) {
	 out<<" CORRECT ----------\n\n\n\n\n";
	 ++tp;
	 p.push_back(tp/(i+1.0));
	 r.push_back(tp/(double)ieg.size());
	 ++att;
	 ++corr;
      }
      else {
	 if (ieg[i].conf >= 0) {
	    out<<" INCORRECT ----------\n\n\n\n\n";
	    p.push_back(tp/(i+1.0));
	    r.push_back(tp/(double)ieg.size());
	    ++att;
	    ++incorr;
	 }
	 else out<<" INCOMPLETE ----------\n\n\n\n\n";
      }
   }
   
   // interpolate precision
   for (int i=p.size()-2; i>=0 ; --i) { 
      if (p[i+1] > p[i]) p[i] = p[i+1];
   }

   double bf=0, pf, rf;
   for (int i=0; i<p.size(); ++i) {
      double f = 2*p[i]*r[i]/(p[i]+r[i]);
      if (f > bf) { 
	 bf = f;
	 pf = p[i];
	 rf = r[i];
      }
   }
   

   out<<"\nTotal: "<<ieg.size()<<"\n";
   out<<"Attempted: "<<att<<"\n";
   out<<"Correct: "<<corr<<"\n";
   out<<"Incorrect: "<<incorr<<"\n";
   if (p.size() > 0) {
      out<<"Precision: "<<p[p.size()-1]*100<<"\n";
      out<<"Recall: "<<r[r.size()-1]*100<<"\n";
      out<<"Best F-measure: "<<bf*100<<" at P: "<<pf*100<<" and R: "<<rf*100<<"\n\n";
   }
   else {
      out<<"Precision: undefined\n";
      out<<"Recall: 0\n";
   }

   out.close();

   cout<<"\nTotal: "<<ieg.size()<<"\n";
   cout<<"Attempted: "<<att<<"\n";
   cout<<"Correct: "<<corr<<"\n";
   cout<<"Incorrect: "<<incorr<<"\n";
   if (p.size() > 0) {
      cout<<"Precision: "<<p[p.size()-1]*100<<"\n";
      cout<<"Recall: "<<r[r.size()-1]*100<<"\n";
      cout<<"Best F-measure: "<<bf*100<<" at P: "<<pf*100<<" and R: "<<rf*100<<"\n\n";
   }
   else {
      cout<<"Precision: undefined\n";
      cout<<"Recall: 0\n";
   }

   ofstream pr("pr");
   for (int i=0; i<p.size(); ++i) {
      pr<<r[i]*100<<" "<<p[i]*100<<"\n";
   }
   pr.close();
}

int main(int argc, char **argv) {

   if (argc < 4) {
      cout<<"USAGE: eval correct_file output_file result_file\n";
      exit(1);
   }

   // read productions
   readGr("gr");
   mrlParser mp(prods,START);

   test(argv[1], argv[2], argv[3], mp);
}
