#include <fstream>
#include <cassert>
#include <string>
#include <map>
#include <functional>
#include <algorithm>

#include "util.h"
#include "parse.h"
#include "production.h"
#include "example.h"
#include "pn.h"
#include "multEg.h"
#include "semanticParser.h"

using namespace std;

string START;

// parameters
bool OUT=false; // trace or not
double THR1=0, THR2=0; // confidence thresholds for including an example, 
                          // first and subsequent iterations 


int KERNEL; // type of kernel: string, tree or combined
double TREE_LAMBDA=0.2; // tree kernel parameter
int STR_KERNEL_LEN=3; // string kernel parameter
double LAMBDA=0.25; // string kernel parameter
string fold="";
string aout="";
string outFile;

int MAX_MEM=1000000; // maximum storage to avoid some recomputations (memory vs time trade-off)
int MAX_MEM2=10000; 

int BEAM=1; // 20
double LIM=0.05; // 0.05 
//int SPAN_LIM=6; // 6
int ITR=2; // 2
int ANN;  // controls the noise level


#define STRING 0
#define TREE 1
#define COMB 2

// global variables
vector <production> prods; // all the productions
vector <vector <string> > synProds; // syntactic productions

// maps to make string computations efficient 
map <string, int> s2n; // mapping from strings to numbers
vector <string> n2s; // reverse mapping
map <string, int> nonterm; // map from non-terminals to indices

vector <vector <int> > nli;  // NL sentences 
vector <ctree> tnli; // trees
vector <double> sknli; // self kernels 
int divide=1;

void writeUsage() {
   cerr<<"USAGE: ./main -train trainFile -test testFile -out outFile\n"; 
}

void processArgs(int argc, char **argv, string &trainFile, string &testFile, string &useDir, string &saveDir) {
   if (argc < 4) {
      writeUsage();
      exit(1);
   }
   int i=1;
   string k;
   KERNEL = STRING;
   trainFile = "";
   testFile = "";
   useDir = "";
   saveDir = "";
   bool od=false;
   while (i<argc) {
      string a = argv[i];
      if (a=="-train") {
	 if (useDir != "") {
	    cerr<<"Incorrect arguments, -train and -use options catn't go together.\n";
	    writeUsage();
	    exit(1);
	 }
	 else {
	    ++i;
	    trainFile = argv[i];
	    ++i;
	 }
	 continue;
      }
      if (a=="-test") {
	 ++i;
	 testFile = argv[i];
	 ++i;
	 continue;
      }
      if (a=="-fold") {
         ++i;
         fold = argv[i];
	 ++i;
	 continue;
      }
      if (a=="-out") {
         ++i;
         aout = argv[i];
	 ++i;
	 continue;
      }
      if (a=="-trace") {
	 ++i;
	 OUT = true;
	 continue;
      }
      if (a=="-thresh1") {
	 ++i;
	 THR1 = atof(argv[i]);
	 ++i;
	 continue;
      }
      if (a=="-thresh2") {
	 ++i;
	 THR2 = atof(argv[i]);
	 ++i;
	 continue;
      }
      cerr<<"Illegal argument: "<<argv[i]<<"\n";
      writeUsage();
      exit(1);
   }
   if (trainFile != "") cout<<"trainFile="<<trainFile<<"\n";
   if (testFile != "") cout<<"testFile="<<testFile<<"\n";

   if (aout != "") {
      outFile = aout;
   }
}


void readGr(const string &file) {
   // read formal grammar
   ifstream in(file.c_str());

   map<string, int>::iterator itr;

   nonterm[""] = 0;

   string s = "";
   while (in && s != "<start>") {
      in>>s;
   }
   assert(in);
   in>>START;
   in>>s;
   assert(s == "</start>");
   
   while (in) {
      production tmp;
      if (tmp.read(in)) {

	 itr = nonterm.find(tmp.getLhs());
	 if (itr==nonterm.end()) {
	    int si = nonterm.size();
	    nonterm[tmp.getLhs()] = si;
	 }

	 if (tmp.isConst) {
	    tmp.pn = nonterm[tmp.getLhs()];
	 }
	 else {
	    tmp.pn = prods.size();
	 }
	 if (tmp.pat.size() > 0) {
	    tmp.isConst = true;
	    tmp.learnable = false;
	 }
	 else {
	    tmp.isConst = false;
	    tmp.learnable = true;
	 }
	 prods.push_back(tmp);
      }
   }
   in.close();

   // if more than two productions with the same constant pattern 
   for (int i=0; i<prods.size(); ++i) {
      prods[i].dupl = false;
      for (int j=0; j<prods.size(); ++j) {
	 if (i != j) {
	    if (prods[i].pat==prods[j].pat) { // later take reorders into account
	       prods[i].dupl = true;
	       break;
	    }
	 }
      }
   }
}

void readTrainExamples(vector <example> &trainEg, const string &filename, const mrlParser &mp) {
   ifstream in(filename.c_str());
   assert(in);

  // read nl and clang examples
   bool b = true;
   while (b) {
      int id;
      vector<std::string> nl1, nl;
      vector <clang> p1;
      double conf;
      b = readExample(in, id, nl1, p1, conf, mp);

      if (b) {
         nGramize(nl1, nl, 1);
	 trainEg.push_back(example(nl, p1, id, prods.size(), conf));
      }
   }
   in.close();
}


void readTestExamples(vector <example> &testEg, const string &filename, const mrlParser &mp) {
   ifstream in(filename.c_str());
   assert(in);

  // read nl and clang examples
   bool b = true;
   while (b) {
      int id;
      vector<std::string> nl1, nl;
      ctree syn, augsyn, mrapt;
      vector <clang> p1;
      b = readExample(in, id, nl1, p1, mp);

      //removeStopWords(nl1);

      if (b) {
         nGramize(nl1, nl, 1);
	 testEg.push_back(example(nl, id, prods.size(), syn, augsyn, mrapt));
      }
   }
   in.close();
}

void replaceConstWords(vector <example> &eg) {
   for (int t=0; t<eg.size(); ++t) {
      eg[t].likelyConsts.resize(prods.size(), false);
      eg[t].orig_nl = eg[t].nl;
      eg[t].orig_nli = eg[t].nli;
      if (eg[t].nl.size() > 20) eg[t].skip = true;
      else eg[t].skip = false;
      vector <string> &s = eg[t].nl;
      for (int p=0; p<prods.size(); ++p) {
	 if (! prods[p].learnable) {
	    for (int i=0; i<prods[p].pat.size(); ++i) {
	       for (int j=0; j<=(int)(s.size()-prods[p].pat[i].size()); ++j) { 
	          bool r = true;
	          for (int e=0; e<prods[p].pat[i].size(); ++e) {
		     if (prods[p].pat[i][e] != s[j+e]) {
		        r = false;
		     }
	          }
	          if (r) {
		     eg[t].skip = false;
		     eg[t].likelyConsts[p] = true;
		     if (!prods[p].dupl) { // replace only when there is no ambiguity 
			for (int e=0; e<prods[p].pat[i].size(); ++e) {
			   if (e==0) {
			      s[j+e] = prods[p].getLhs();
			      eg[t].nli[j+e] = getS2N(prods[p].getLhs());
			   }
			   else {
			      s[j+e] = "??";
			      eg[t].nli[j+e] = getS2N("??");
			   }
			}
		        for (int pp=0; pp<prods.size(); ++pp) { 
			   // if the const already covers then no need to learn 
			   if (prods[pp].learnable && prods[pp].getLhs()==prods[p].getLhs() && prods[pp].getRhs()==prods[p].getRhs()) {
			      eg[t].y[pp] = -1;
			   }
		        }
		     }
	          }
	       }
	    }
	 }
      }
   }
}



void readExamples(vector <example> &trainEg, const string &trainFile, vector <example> &testEg, const string &testFile, const mrlParser &mp) {
   readTrainExamples(trainEg, trainFile,mp);
   replaceConstWords(trainEg);
   readTestExamples(testEg,  testFile,mp);
   replaceConstWords(testEg);
   for (int i=0; i<testEg.size(); ++i) {
      testEg[i].trainEg = false;
   }
}

void writeOut(const vector <example> &testEg, const string &outFile) {
   ofstream out(outFile.c_str());
   for (int i=0; i<testEg.size(); ++i) {
      out<<"\n\n";
      testEg[i].c.writeParse(out);
      testEg[i].write(out);
      testEg[i].writePartition(out);
   }
   out.close();
}

bool changeClang(clang &c, const string &lhs, const vector <string> &rhs, const string &t, const string &nt) {
   bool r=false;
   if (c.s==lhs && c.childSize()==rhs.size()) {
      bool f=true;
      for (int i=0; i<c.childSize(); ++i) {
	 if (c.child[i]==NULL || rhs[i]!=c.child[i]->s) f=false;
      }
      if (f) {
	 r=true;
         for (int i=0; i<c.childSize(); ++i) {
	    if (c.child[i]->s==t) {
	       c.child[i]->s = nt;
	       c.child[i]->makeChild(clang(t), 0);
	       break;
	    }
	 }
      }
   }
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL) {
	 r = r || changeClang(*c.child[i],lhs,rhs,t,nt);
      }
   }
   return r;
}

void createProd(int p, const string &t, const string &nt, const vector <production> &prods, vector <production> &newProds, vector <example> &trainEg) {
   // forms new productions newProds, replacing from prods the pth production with terminal 
   // t (only the first occurrence) with a production with non-terminal nt and 
   // adding a new production nt -> t if it does not already exist,
   // also changes the training MR parse trees to this effect

   newProds = prods;

   vector <string> rhs = newProds[p].getRhs();
   vector <string> oldRhs = rhs;
   for (int i=0; i<rhs.size(); ++i) {
      if (rhs[i]==t) {
	 rhs[i] = nt;
	 break;
      }
   }
   newProds[p].putRhs(rhs);


   bool add=false;
      
   production newp(nt,vector <string>(1,t),0); // add production "nt -> t"
   cout<<newp<<"\n";
   newp.learnable = true;
   newp.isConst = false;
   int ap;
   for (ap=0; ap<newProds.size(); ++ap) {
      if (newp==newProds[ap]) break;
   }
   if (ap==newProds.size()) {
      newProds.push_back(newp); // add production "nt -> t"
      add=true;
      if (nonterm.find(nt)==nonterm.end()) {
	 int si = nonterm.size();
	 nonterm[nt] = si;
      }
   }
   for (int i=0; i<trainEg.size(); ++i) {
      bool ch = changeClang(trainEg[i].c, prods[p].getLhs(), oldRhs, t, nt);
      if (ch) {
	 trainEg[i].c.writeParse(cout);
      }
      int a;
      if (ch) a=1;
      else a = -1;
      if (add) {
	 trainEg[i].y.push_back(a);
      }
      else {
	 if (ch) trainEg[i].y[ap] = a;
      }
      if (ch) {
	 for (int p=0; p<trainEg[i].y.size(); ++p) {
	    if (trainEg[i].y[p]==1) {
	       cout<<newProds[p]<<"\n";
	    }
	 }
      }
   }
}

extern vector <hash_map <int, double> > h;
extern vector <hash_map <int, double> > hn;
int main(int argc, char **argv) {

   string trainFile, testFile, useDir, saveDir;
   processArgs(argc, argv, trainFile, testFile, useDir, saveDir);

   // read grammar
   readGr("gr");

   mrlParser mp(prods,START);

   // read examples
   vector <multEg> megs;
   vector <example> testEg;
   cout<<"Reading training examples..\n";
   readMultEgs(megs, trainFile, mp);
   cout<<"done.\n";

   // write training examples
   //writeVector(cout, megs);
   //exit(1);

   cout<<"Reading test examples..\n";
   readTestExamples(testEg,  testFile, mp);
   cout<<"done.\n";
   replaceConstWords(testEg);
   for (int i=0; i<testEg.size(); ++i) {
      testEg[i].trainEg = false;
   }

   // training and testing
   mult_train_test(megs, testEg);

   // write output
   writeOut(testEg, outFile);
}
