#ifndef _CTREE_H
#define _CTREE_H

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cassert>
#include <cstdio>

using namespace std;

extern double TREE_LAMBDA; // tree kernel parameter

class ctree;
 
class NT { // representation of node used in kernel computation
   public:
      int nodeId, rootProd;
      ctree *ptr;
      bool operator < (const NT &other) const {
	 return rootProd < other.rootProd;
      }
};


class ctree {

   public:
   vector <NT> nt;
   bool ntComputed;
   
   string root;
   vector <ctree*> child;
   
   int rootProd, rootMRProd;
   int n; // a number for each leaf (increases from left to right)
   int sleaf, eleaf; // numbers of start and end leaves
   int nodeId; 
   int size; // size of the tree, i.e. total # of nodes

   bool preTerminal; // pre-terminal
   bool preTermComputed; 
   int d; // depth

   bool readHelp(ifstream &in);
   bool readHelp(FILE *in);

   int findSize();
   int numberLeaves(int i);
   int assignNodeIds(int i);
   void assignDepth(int i);
   void getProds();
   void init();  // initializes the tree

   double sk; // self kernel value 

   bool covers(int s, int e) const {
      return sleaf <= s && eleaf >= e;
   }

   ctree buildCover(int s, int e) const;

   void collapseOneChild() {
      // collapses all nodes with exactly one child
      while (child.size()==1 && root ==  child[0]->root) {
	 root = child[0]->root;
	 n = child[0]->n;
	 child = child[0]->child;
      }
      for (int i=0; i<child.size(); ++i) child[i]->collapseOneChild();
   }

   void assignFrom(const ctree&t) {
      root = t.root;
      n = t.n;
      sleaf = t.sleaf;
      eleaf = t.eleaf;
      rootProd = t.rootProd;
      rootMRProd = t.rootMRProd;
      nodeId = t.nodeId;
      size = t.size;
      sk = t.sk;
      d = t.d;
      preTerminal = t.preTerminal;
      preTermComputed = t.preTermComputed;
      ntComputed = false;
   }


   public:
   
   ctree () {
      root = "";
      size = 0;
      sk = -1;
      ntComputed = false;
      preTermComputed = false;
   }
   ctree (const string &s) {
      root = s;
      size = 1;
      sk = -1;
      ntComputed = false;
      preTermComputed = false;
   }
   void ntComp() {
      sort(nt.begin(), nt.end());
      ntComputed = true;
   }
   bool ifNtComp() const {
      return ntComputed;
   }
   bool read(ifstream &in);
   bool read(FILE *in);
   ostream& write(ostream &out) const;
   ostream& writeMR(ostream &out) const;

   bool empty() const {
      return root=="";
   }

   bool preTerm();

   int giveProd () const {
      return rootProd;
   }
   int treeSize() const {
      return size;
   }

   ~ctree () {
      for (int i=0; i<child.size(); ++i) {
         if (child[i] != NULL) {
	    delete child[i];
         }
      }
   }
    
   ctree (const ctree &t) {
      assignFrom(t);
      for (int i=0; i<t.child.size(); ++i) {
	 if (t.child[i] != NULL) child.push_back(new ctree(*(t.child[i])));
      }
   }
   void operator=(const ctree &t) {
      assignFrom(t);
      for (int i=0; i<child.size(); ++i) {
         if (child[i] != NULL) delete child[i];
      }
      child.clear();
      for (int i=0; i<t.child.size(); ++i) {
         if (t.child[i] != NULL) child.push_back(new ctree(*(t.child[i])));
      }
   }
   bool operator == (const ctree &other) const {
      if (child.size() != other.child.size()) return false;
      if (root != other.root) return false;
      for (int i=0; i<child.size(); ++i) {
	 if (*child[i] != *other.child[i]) return false;
      }
      return true;
   }
   bool operator != (const ctree &other) const {
      return !(operator==(other));
   }

   int childSize() const {
      return child.size();
   }
   ctree giveChild(int i) const{
      return *child[i];
   }
   string giveRoot() const {
      return root;
   }

   int giveNodeId() const {
      return nodeId;
   }

   ctree leafCover(int s, int e) const; // returns subtree covering the leaves from s to e

   void leafString(vector <string> &ls) const; // its string of leaves 

   void posEgs(vector <int> &p, vector <int> &sl, vector <int> &el); // positive examples from the mrapt (start and end leaves)
   void posEgs_helper(vector <int> &p, vector <int> &sl, vector <int> &el) const; // positive examples from the mrapt (start and end leaves)

   string firstAncestor(int s, int e) const;

};

inline ostream& operator << (ostream& lhs, const  ctree& rhs) {
   return rhs.write(lhs);
}
   
double kernelUn(ctree &, ctree &);
double selfKernel(ctree &);
double kernel(ctree &, ctree &);

#endif
