// This program implements the tree kernel from [Collins & Duffy, NIPS'01]. 

/* 
 treeKernel() is a templated function which expects the
 tree data structure to provide the following functions:
 giveProd(): returns the production number at the node (unique for a production)
 childSize(): number of children at the node
 giveChild(i): return the ith child
 giveRoot(): gives the label of the node
*/

#ifndef _TREEKERNEL_H
#define _TREEKERNEL_H

#include <iostream>
#include <vector>
#include <string>
#include <cassert>
#include "production.h"

using namespace std;

extern vector <vector <string> > synProds; // syntactic productions
extern double TREE_LAMBDA; // tree kernel parameter

class nodeInfo {
   public:
      string s;
      bool complete;
      int prod; // production number
      vector <int> childIndex; // indices of children
      nodeInfo () {
	 complete = false;
      }
      nodeInfo(const string &ss){
	 s=ss;
	 complete = false;
      }
      ostream &write(ostream &out) const {
	 if (complete) {
	 out<<prod<<": \n"; 
	 for (int e=0; e<synProds[prod].size(); ++e) out<<synProds[prod][e]<<" ";
	 out<<" ";
	 for (int i=0; i<childIndex.size(); ++i) {
	    out<<childIndex[i]<<" ";
	 }
	 out<<"\n";
	 }
	 return out;
      }
};
inline ostream& operator << (ostream& lhs, const nodeInfo& rhs) {
   return rhs.write(lhs);
}

template <typename Tree>
void getNodeInfo(const Tree &t, vector <nodeInfo> &ni, int ind=0) {
   // ni will store prods of all the nodes and the indices of children
   // in this vector
   if (ni.size() <= ind) ni.resize(ind+1);
   ni[ind].prod = t.giveProd();
   ni[ind].s = t.giveRoot();
   for (int i=0; i<t.childSize(); ++i) {
      Tree gc = t.giveChild(i);
      if (gc.childSize() > 0) {
         ni.push_back(nodeInfo());
         ni[ind].childIndex.push_back(ni.size()-1);
         getNodeInfo(gc, ni, ni.size()-1);
      }
      else {
         ni[ind].childIndex.push_back(-1);
      }
   }
   ni[ind].complete = true;
}

/*
template <typename Tree>
void getNodeInfoInd(Tree &t, vector <nodeInfo> &ni, int ind=0) {
   // ni will store prods of all the nodes and the indices of children
   // in this vector
   if (ni.size() <= ind) ni.resize(ind+1);
   ni[ind].prod = t.giveProd();
   ni[ind].s = t.giveRoot();
   t.ind = ind;
   for (int i=0; i<t.childSize(); ++i) {
      if (t.child[i]->childSize() > 0) {
         ni.push_back(nodeInfo());
         ni[ind].childIndex.push_back(ni.size()-1);
         getNodeInfoInd(*t.child[i], ni, ni.size()-1);
      }
      else {
         ni[ind].childIndex.push_back(-1);
      }
   }
   ni[ind].complete = true;
}*/


double computeC(int i, int j,  vector <vector <double> > &c, vector <vector <double> > &done, const  vector <nodeInfo> &ni1,  const vector <nodeInfo> &ni2);

double treeKernelUn(const vector <nodeInfo> &ni1, const vector <nodeInfo> &ni2);
double treeKernel(const vector <nodeInfo> &ni1, const vector <nodeInfo> &ni2);

/*
template <typename Tree>
void putcst(Tree &t, const vector <double> &cst) {
   if (t.childSize() > 0) {
      t.cst = cst[t.ind];
      for (int i=0; i<t.childSize(); ++i) putcst(*t.child[i], cst);
   }
}

template <typename Tree>
double allCommonSubtrees(Tree &t1, const Tree &t2) {
   // each node in t1 will store the number of common subtrees rooted at it that are present in t2
   vector <nodeInfo> ni1, ni2;
   getNodeInfoInd(t1, ni1);
   getNodeInfo(t2, ni2);
   
   vector <vector <double> > c(ni1.size(), vector <double> (ni2.size(), 0));
   vector <vector <double> > done(ni1.size(), vector <double> (ni2.size(), false));
   vector <double> cst(ni1.size(), 0);

   for (int i=0; i<ni1.size(); ++i) {
      for (int j=0; j<ni2.size(); ++j) {
	 cst[i] += computeC(i, j, c, done, ni1, ni2);
      }
   }
   putcst(t1, cst);
}*/


template <typename Tree>
double computeC(const Tree &t1, const Tree &t2, vector <vector <double> > &c,  vector <vector <double> > &done) {
   int i = t1.giveNodeId();
   int j = t2.giveNodeId();
   
   if (done[i][j]) return c[i][j];

   if (t1.giveProd() != t2.giveProd()) {
      // nodes have different productions
      c[i][j] = 0;
      done[i][j] = true;
   }
   else {
      if (t1.childSize()==0 || t2.childSize()==0) {
	 // nodes are leaves
	 c[i][j] = TREE_LAMBDA;
	 done[i][j] = true;
      }
      else {
	 // recursive case
	 double r=TREE_LAMBDA;
	 assert(t1.childSize()==t2.childSize());
	 for (int e=0; e<t1.childSize(); ++e) {
	    r *= 1 + computeC(*t1.child[e]Index[e], t2.child[e], c, done);
	 }
	 c[i][j] = r;
	 done[i][j] = true;
      }
   }
   return c[i][j];
}

template <typename Tree>
void computeAllNodesC(const Tree &t1, const Tree &t2, vector <vector <double> > &c,  vector <vector <double> > &done) {
   computeC(t1, t2);
   for (int e=0; e<t2.childSize(); ++e) {
      computeAllNodesC(t1, *t2.child[e]);
   }
}

template <typename Tree>
double treeKernelUn_new(const Tree &t1, const Tree &t2) {
   vector <vector <double> > c(t1.treeSize(), vector <double> (t2.treeSize(), 0));
   vector <vector <double> > done(t1.treeSize(), vector <double> (t2.treeSize(), false));

   double r=0;
   for (int i=0; i<ni1.size(); ++i) {
      for (int j=0; j<ni2.size(); ++j) {
	 r += computeC(i, j, c, done, ni1, ni2);
      }
   }
   return r;
}
   
template <typename Tree>
double treeKernelUn(const Tree &t1, const Tree &t2) {
   // unnormalized tree kernel
   //cout<<t1<<"\n";
   //cout<<t2<<"\n--------------\n";
   vector <nodeInfo> ni1, ni2;
   getNodeInfo(t1, ni1);
   getNodeInfo(t2, ni2);
   return treeKernelUn(ni1, ni2);
}


template <typename Tree>
double treeKernel(const Tree &t1, const Tree &t2) {
   // normalized tree kernel
   //cout<<t1<<"\n";
   //cout<<t2<<"\n----------------\n";
   vector <nodeInfo> ni1, ni2;
   getNodeInfo(t1, ni1);
   getNodeInfo(t2, ni2);
   return treeKernel(ni1, ni2);
}

#endif
